# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = "pangeo-forge-runner"
copyright = "2022, YuviPanda"
author = "YuviPanda"

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    "sphinx.ext.autodoc",
    "autodoc_traits",
    "sphinx_copybutton",
    "myst_parser",
]

templates_path = ["_templates"]
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]

source_suffix = [".rst", ".md"]

# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = "pydata_sphinx_theme"
html_static_path = ["_static"]

html_theme_options = {
    "icon_links": [
        {
            "name": "GitHub",
            "url": "https://github.com/pangeo-forge/pangeo-forge-runner",
            "icon": "fa-brands fa-github",
        },
    ],
}



import apache_beam as random_name_because_why_not
from pangeo_forge_recipes.patterns import pattern_from_file_sequence

pattern = pattern_from_file_sequence(["hi.nc"])

recipe = random_name_because_why_not.Create(pattern.prune().items())



import base64
import os
import secrets
import signal
import socket
import subprocess
import tempfile

import pytest


# Stolen from https://stackoverflow.com/a/28950776
@pytest.fixture(scope="session")
def local_ip():
    """
    Return IP of current machine

    Hopefully, this is resolveable by both code running on the machine
    as well as whatever kubernetes cluster is being used to run tests.
    """
    s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    s.settimeout(0)
    try:
        # doesn't even have to be reachable
        s.connect(("10.254.254.254", 1))
        IP = s.getsockname()[0]
    except Exception:
        IP = "127.0.0.1"
    finally:
        s.close()
    return IP


@pytest.fixture(scope="session")
def minio(local_ip):
    """
    Start a temporary minio instance & return values used to connect to it

    yields a dictionary with the following keys:
    - endpoint - the HTTP endpoint to use to talk to the minio API
    - username - the minio root username (aka 'key' in S3 parlance)
    - password - the minio root password
    """
    username = secrets.token_hex(16)
    password = secrets.token_hex(16)
    address = f"{local_ip}:19555"
    endpoint = f"http://{address}"

    env = os.environ.copy()
    env.update({"MINIO_ROOT_USER": username, "MINIO_ROOT_PASSWORD": password})
    with tempfile.TemporaryDirectory() as d:
        proc = subprocess.Popen(["minio", "server", d, "--address", address], env=env)

        yield {"endpoint": endpoint, "username": username, "password": password}

        # Cleanup minio server during teardown
        proc.send_signal(signal.SIGTERM)
        proc.wait()

        assert proc.returncode == 0


@pytest.fixture(scope="session")
def minio_service():
    cmd = [
        "kubectl",
        "get",
        "service/minio-service",
        "-o=jsonpath='{.spec.clusterIP}:{.spec.ports[0].port}'",
    ]
    proc = subprocess.run(cmd, capture_output=True, text=True)
    assert proc.returncode == 0
    svc_address = proc.stdout.strip('"').strip("'")
    endpoint = f"http://{svc_address}"

    cmd = [
        "kubectl",
        "get",
        "secret/minio-secrets",
        "-o=jsonpath='{.data.MINIO_ACCESS_KEY}'",
    ]
    proc = subprocess.run(cmd, capture_output=True, text=True)
    assert proc.returncode == 0
    myaccesskey = proc.stdout
    myaccesskey = base64.b64decode(myaccesskey).decode()

    cmd = [
        "kubectl",
        "get",
        "secret/minio-secrets",
        "-o=jsonpath='{.data.MINIO_SECRET_KEY}'",
    ]
    proc = subprocess.run(cmd, capture_output=True, text=True)
    assert proc.returncode == 0
    mysecretkey = proc.stdout
    mysecretkey = base64.b64decode(mysecretkey).decode()

    # enter
    yield {"endpoint": endpoint, "username": myaccesskey, "password": mysecretkey}

    # exit
    return


def pytest_addoption(parser):
    parser.addoption("--flinkversion", action="store", default="1.16")
    parser.addoption("--pythonversion", action="store", default="3.9")
    parser.addoption("--beamversion", action="store", default="2.47.0")


@pytest.fixture
def flinkversion(request):
    return request.config.getoption("--flinkversion")


@pytest.fixture
def pythonversion(request):
    return request.config.getoption("--pythonversion")


@pytest.fixture
def beamversion(request):
    return request.config.getoption("--beamversion")



import shutil
import subprocess
import uuid

import pytest
from traitlets import TraitError

from pangeo_forge_runner.bakery.dataflow import DataflowBakery


@pytest.mark.parametrize(
    "property, attr_name, new_value, should_set_as_default",
    (
        ["project", "project_id", str(uuid.uuid4()), True],
        [
            "account",
            "service_account_email",
            "bot@project.iam.gserviceaccount.com",
            True,
        ],
        ["account", "service_account_email", "user@university.edu", False],
    ),
)
def test_default_gcp_props(property, attr_name, new_value, should_set_as_default):
    """
    Test using `gcloud` to determine currently active properties
    """
    # We want `gcloud` to actually be present in the environment
    assert shutil.which("gcloud")

    # We fetch the current project (if it exists) to make sure we restore it after our test
    current_value = None

    proc = subprocess.run(
        ["gcloud", "config", "get-value", property], stdout=subprocess.PIPE
    )
    if proc.returncode == 0:
        current_value = proc.stdout.decode().strip()

    try:
        # Set stdin to /dev/null so `gcloud` doesn't ask us 'why are you setting project to something you can not access?'
        subprocess.run(
            ["gcloud", "config", "set", property, new_value],
            check=True,
            stdin=subprocess.DEVNULL,
        )
        dfb = DataflowBakery()
        if should_set_as_default:
            assert getattr(dfb, attr_name) == new_value
        else:
            assert getattr(dfb, attr_name) is None
    finally:
        if current_value:
            subprocess.run(["gcloud", "config", "set", property, current_value])


def test_temp_gcs_location_validation():
    dfb = DataflowBakery()
    with pytest.raises(TraitError):
        dfb.temp_gcs_location = "This Should be an Error"

    dfb.temp_gcs_location = "gs://this-should-not-error"


def test_pipelineoptions():
    """
    Quickly validate some of the PipelineOptions set
    """
    dfb = DataflowBakery()
    dfb.project_id = "hello"
    dfb.region = "us-west1"
    dfb.machine_type = "n1-standard-2"
    dfb.use_public_ips = True
    dfb.temp_gcs_location = "gs://something"

    po = dfb.get_pipeline_options(
        "job", "some-container:some-tag", {"requirements_file": "/tmp/some-file"}
    )
    opts = po.get_all_options()
    assert opts["project"] == "hello"
    assert opts["use_public_ips"]
    assert opts["temp_location"] == "gs://something"
    assert opts["machine_type"] == "n1-standard-2"
    assert opts["region"] == "us-west1"
    assert opts["experiments"] == ["use_runner_v2"]
    assert opts["save_main_session"]
    assert opts["pickle_library"] == "cloudpickle"
    assert opts["sdk_container_image"] == "some-container:some-tag"
    assert opts["job_name"] == "job"
    assert opts["runner"] == "DataflowRunner"
    assert opts["requirements_file"] == "/tmp/some-file"


def test_required_params():
    dfb = DataflowBakery()
    dfb.project_id = None
    dfb.temp_gcs_location = "gs://test"

    with pytest.raises(ValueError):
        dfb.get_pipeline_options("test", "test", {})

    dfb.project_id = "something"
    dfb.temp_gcs_location = None

    with pytest.raises(ValueError):
        dfb.get_pipeline_options("test", "test", {})


def test_missing_gcloud(mocker):
    def find_nothing_which(cmd):
        return False

    mocker.patch("shutil.which", find_nothing_which)

    dfb = DataflowBakery()
    assert dfb.project_id is None


def test_dataflow_prime():
    """
    Validate that machine_type is not set when dataflow prime is enabled
    """
    dfb = DataflowBakery()
    dfb.use_dataflow_prime = True
    dfb.temp_gcs_location = "gs://something"

    po = dfb.get_pipeline_options(
        "job", "some-container:some-tag", {"requirements_file": "/tmp/some-file"}
    )
    opts = po.get_all_options()
    assert opts["machine_type"] is None
    assert opts["dataflow_service_options"] == ["enable_prime"]



import jsonschema
from traitlets import Dict, HasTraits, List, TraitError, Unicode, validate

recipes_field_per_element_schema = {
    "type": "object",
    "properties": {
        "id": {"type": "string"},
        "object": {"type": "string"},
        "dict_object": {"type": "string"},
    },
    "oneOf": [
        {"required": ["id", "object"]},
        {"required": ["dict_object"]},
    ],
}


class MetaYaml(HasTraits):
    """Schema for the ``meta.yaml`` file which must be included in each feedstock directory.
    Only the ``recipes`` field is strictly required for ``pangeo-forge-runner`` to function.
    All other fields are recommended but not required.
    """

    def __init__(self, recipes=None, **kwargs):
        """The only required field is ``recipes``, so we put it explicitly in the init
        signature to ensure it is not omitted, as demonstrated in:
        https://github.com/ipython/traitlets/issues/490#issuecomment-479716288
        """
        super().__init__(**kwargs)
        self.recipes = recipes

    @validate("recipes")
    def _validate_recipes(self, proposal):
        """Ensure the ``recipes`` trait is not passed as an empty container and that
        each element of the field contains all expected subfields.
        """
        if not proposal["value"]:
            raise TraitError(
                f"The ``recipes`` trait, passed as {proposal['value']}, cannot be empty."
            )

        if isinstance(proposal["value"], list):
            for recipe_spec in proposal["value"]:
                try:
                    jsonschema.validate(recipe_spec, recipes_field_per_element_schema)
                except jsonschema.ValidationError as e:
                    raise TraitError(e)
        return proposal["value"]

    title = Unicode(
        allow_none=True,
        help="""
        Title for this dataset.
        """,
    )
    description = Unicode(
        allow_none=True,
        help="""
        Description of the dataset.
        """,
    )
    recipes = List(
        Dict(),
        help="""
        Specifies the deployable Python objects to run in the recipe module.
        If the recipes are assigned to their own Python variable names,
        should be of the form:

        ```yaml
        recipes:
          - id: "unique-identifier-for-recipe"
            object: "recipe:transforms"
        ```
        
        Alternatively, if the recipes are values in a Python dict:

        ```yaml
        recipes:
          - dict_object: "recipe:name_of_recipes_dict"
        ```
        """,
    )
    provenance = Dict(
        allow_none=True,
        help="""
        Dataset provenance information including provider, license, etc.
        """,
        per_key_traits={
            "providers": List(
                Dict(
                    per_key_traits={
                        "name": Unicode(),
                        "description": Unicode(),
                        "roles": List(),  # TODO: enum
                        "url": Unicode(),
                    },
                ),
            ),
            "license": Unicode(),  # TODO: guidance on suggested formatting (enum?)
        },
    )
    maintainers = List(
        Dict(
            per_key_traits={
                "name": Unicode(help="Full name of the maintainer."),
                "orcid": Unicode(help="Maintainer's ORCID ID"),  # TODO: regex
                "github": Unicode(help="Maintainer's GitHub username."),
            },
        ),
        allow_none=True,
        help="""
        Maintainers of this Pangeo Forge feedstock.
        """,
    )



"""
Handle the plugin system for injections.

There are three parts of injections:

1. An "injection spec", provided by other installed packages (such as
   pangeo_forge_recipes, pangeo_forge_cmr, etc). This specifies what
   *values* exactly will be injected as args for *which* callables.
   It is in the form of a dictionary, and looks like this:
   ```
   {
     "<callable-1-name>": {
       "<argument-1-name>": "<value-spec>",
       "<argument-2-name>": "<value-spec>"
     },
     "<callable-2-name>": {
       "<argument-1-name>": "<value-spec>",
       "<argument-2-name>": "<value-spec>"
     }
   }
   ```

   `<value-spec>` specifies what value should be injected. Currently
   supported are two strings:
   1. `TARGET_STORAGE` - Storage that a pipeline *output* should be written to.
       Will be passed as a `pangeo_forge_recipes.storage.FSSpecTarget` object.
   2. `INPUT_CACHE_STORAGE` - (Optional) Storage used for caching inputs.
       Will be passed as a `pangeo_forge_recipes.storage.CacheFSSpecTarget` object.


   Additional values may be provided in the future.

   An example is:

   ```
    {
        'StoreToZarr': {
            'target_root': 'TARGET_STORAGE',
        },
        'OpenURLWithFSSpec': {
            'cache': 'INPUT_CACHE_STORAGE'
        }
    }
    ```

   We considered making this into an Enum, but that would have required all
   packages that provide entrypoints also *import* pangeo_forge_runner. This
   was deemed too complicating, and hence raw strings are used.

2. "Injection spec values", calculated by pangeo-forge-runner. This is simply a
   mapping of "<value-spec>" to a specific value that will be injected for
   that "<value-spec>" in this particular run. This might look like:

   ```
   {
     "TARGET_STORAGE": <A `pangeo_forge_recipes.storage.FSSpecTarget` object>,
     "INPUT_CACHE_STORAGE": <A `pangeo_forge_recipes.storage.CacheFSSpecTarget` object>
   }
   ```
3. "Injections", ready to be passed on to the rewriter! This merges (1) and (2),
   and looks like:
   ```
    {
        'StoreToZarr': {
            'target_root': <A `pangeo_forge_recipes.storage.FSSpecTarget` object>
        },
        'OpenURLWithFSSpec': {
            'cache': <A `pangeo_forge_recipes.storage.CacheFSSpecTarget` object>
        }
    }
   ```

   This is what is actually injected into the recipes in the end.
"""
# Use the backported importlib_metadata as we still support Python 3.9
# Once we're on 3.10 we can remove this dependency and use the built in
# importlib.metadata
from importlib_metadata import entry_points
from jsonschema import validate

# Schema for the dictionary returned by injection spec entrypoints
INJECTION_SPEC_SCHEMA = {
    "type": "object",
    # patternProperties to allow arbitrary keys. The first level keys represent
    # callable names.
    "patternProperties": {
        ".+": {
            "type": "object",
            # Second level keys represent attribute names in the callable, and are also arbitray.
            "patternProperties": {
                # Value of the second level keys is restricted to just these two
                ".+": {
                    "type": "string",
                    "enum": ["TARGET_STORAGE", "INPUT_CACHE_STORAGE"],
                }
            },
        }
    },
    "additionalProperties": False,
}


def get_injectionspecs_from_entrypoints():
    """
    Collection injectionspecs from installed packages.

    Looks for entrypoints defined in installed packages with the
    group "pangeo_forge_runner.injections", and calls them all in
    an undefined order. They are expected to return a dict with
    specification of what exactly should be injected where, and then
    merged together.
    """
    injection_specs = {}
    eps = entry_points(group="pangeo_forge_runner.injection_specs")
    for ep in eps:
        specs = ep.load()()
        # FIXME: This throws an exception, but user doesn't know which plugin actually
        # failed validation! provide that information
        validate(specs, schema=INJECTION_SPEC_SCHEMA)
        # FIXME: This is a shallow merge, should be a deep merge instead
        injection_specs |= specs

    if injection_specs == {}:
        # Handle the specific case of pangeo-forge-recipes==0.10.x,
        # which shipped with beam transforms that need injections, but without
        # entrypoint based injection specs.
        injection_specs = {
            "StoreToZarr": {
                "target_root": "TARGET_STORAGE",
            },
            "WriteCombinedReference": {
                "target_root": "TARGET_STORAGE",
            },
            "OpenURLWithFSSpec": {"cache": "INPUT_CACHE_STORAGE"},
        }

    return injection_specs


def get_injections(injection_spec: dict, injection_values: dict) -> dict[str, str]:
    """
    Given an injection_spec and injection_values, provide actual injections
    """
    injections = {}

    for cls, params in injection_spec.items():
        for param, target in params.items():
            if target in injection_values:
                injections.setdefault(cls, {})[param] = injection_values[target]

    return injections



from pathlib import Path

import pytest
from ruamel.yaml import YAML
from traitlets import TraitError

from pangeo_forge_runner import Feedstock

yaml = YAML()

HERE = Path(__file__).parent


def test_recipes_list():
    list_recipe = HERE / "test-recipes/list-recipes/feedstock"
    feed = Feedstock(list_recipe)
    recipes = feed.parse_recipes()
    assert recipes == {"test_1": "test_1", "test_2": "test_2"}

    with open(list_recipe / "meta.yaml") as f:
        meta = yaml.load(f)
        assert meta == feed.get_expanded_meta()


def test_recipes_dict():
    list_recipe = HERE / "test-recipes/dict-recipes/feedstock"
    feed = Feedstock(list_recipe)
    recipes = feed.parse_recipes()
    assert recipes == {"test_1": "test_1", "test_2": "test_2"}

    with open(list_recipe / "meta.yaml") as f:
        meta = yaml.load(f)
        meta["recipes"] = [
            {"id": "test_1", "object": "DICT_VALUE_PLACEHOLDER"},
            {"id": "test_2", "object": "DICT_VALUE_PLACEHOLDER"},
        ]
        assert meta == feed.get_expanded_meta()


def test_recipes_broken():
    list_recipe = HERE / "test-recipes/broken-recipe/feedstock"
    with pytest.raises(TraitError):
        _ = Feedstock(list_recipe)



import ast
from copy import deepcopy
from pathlib import Path
from typing import Optional

from ruamel.yaml import YAML

from .meta_yaml import MetaYaml
from .recipe_rewriter import RecipeRewriter

yaml = YAML()


class Feedstock:
    """
    A Pangeo Forge feedstock
    """

    def __init__(
        self,
        feedstock_dir: Path,
        prune: bool = False,
        callable_args_injections: Optional[dict] = None,
    ):
        """
        feedstock_dir: Path to an existing Feedstock repo
        prune: Set to true if the recipe should be pruned for testing
        callable_args_injections: A dict of callable names (as keys) with injected kwargs as value

        Expects meta.yaml to exist inside in this dir
        """
        self.feedstock_dir = feedstock_dir
        with open(self.feedstock_dir / "meta.yaml") as f:
            self.meta = MetaYaml(**yaml.load(f))

        self.prune = prune
        self.callable_args_injections = (
            callable_args_injections if callable_args_injections else {}
        )

    def _import(self, spec):
        """
        Import & return given object from recipes/ in feedstock_dir

        spec is of form <filename>:<object-name>, similar to what is used
        elsewhere in python.

        Each file is executed only once and cached.
        """
        if not hasattr(self, "_import_cache"):
            self._import_cache = {}

        rewriter = RecipeRewriter(
            prune=self.prune, callable_args_injections=self.callable_args_injections
        )

        module, export = spec.split(":")
        if module not in self._import_cache:
            ns = {**rewriter.get_exec_globals()}
            filename = self.feedstock_dir / f"{module}.py"
            with open(filename) as f:
                # compiling makes debugging easier: https://stackoverflow.com/a/437857
                # Without compiling first, `inspect.getsource()` will not work, and
                # pangeo-forge-recipes uses it to hash recipes!
                recipe_ast = ast.parse(source=f.read(), filename=filename, mode="exec")
                rewritten_ast = rewriter.visit(recipe_ast)
                exec(compile(source=rewritten_ast, filename=filename, mode="exec"), ns)
                self._import_cache[module] = ns

        return self._import_cache[module][export]

    def parse_recipes(self):
        """
        Parse the recipes defined in this feedstock & return them

        *Executes arbitrary code* defined in the feedstock recipes.
        """
        recipes = {}
        for r in self.meta.recipes:
            assert any(
                # MetaYaml schema validation of self.meta ensures that one of these two
                # conditions is true, but just assert anyway, to make sure there are no
                # edge cases that slip through the cracks.
                [
                    key_set == set(r.keys())
                    for key_set in ({"id", "object"}, {"dict_object"})
                ]
            )
            if {"id", "object"} == set(r.keys()):
                recipes[r["id"]] = self._import(r["object"])
            elif {"dict_object"} == set(r.keys()):
                dict_object = self._import(r["dict_object"])
                for k, v in dict_object.items():
                    recipes[k] = v

        return recipes

    def get_expanded_meta(self, drop_none=True) -> dict:
        """
        Return full meta.yaml file, expanding recipes if needed.

        recipes are guaranteed to be a list of dictionaries with 'id' keys.
        'object' keys *may* be present, but not guaranteed.
        """
        meta_copy = deepcopy(self.meta)
        if any(["dict_object" in r for r in self.meta.recipes]):
            if not all(["dict_object" in r for r in self.meta.recipes]):
                raise NotImplementedError(
                    "Mixing explicit recipe objects and dict_objects in a "
                    "single feedstock is not yet supported."
                )
            # We have at least one dict_object, so let's parse the recipes,
            # keeping just the ids, discarding the values - as the values do not
            # actually serialize.
            meta_copy.recipes = [
                # In place of dict values, we add a placeholder string, so that the
                # re-assignment to the MetaYaml schema here will pass validation
                # of the `recipes` field, which requires "id" and "object" fields.
                {"id": k, "object": "DICT_VALUE_PLACEHOLDER"}
                for k, _ in self.parse_recipes().items()
            ]
        return (
            # the traitlets MetaYaml schema will give us empty containers
            # by default, but in most cases lets assume we don't want that
            {k: v for k, v in meta_copy.trait_values().items() if v}
            if drop_none
            else meta_copy.trait_values()
        )



"""
Command to run a pangeo-forge recipe
"""
import hashlib
import os
import re
import string
import time
from importlib.metadata import distributions
from pathlib import Path

import escapism
from apache_beam import Pipeline, PTransform
from traitlets import Bool, Type, Unicode, validate

from .. import Feedstock
from ..bakery.base import Bakery
from ..bakery.flink import FlinkOperatorBakery
from ..bakery.local import LocalDirectBakery
from ..plugin import get_injections, get_injectionspecs_from_entrypoints
from ..storage import InputCacheStorage, MetadataCacheStorage, TargetStorage
from ..stream_capture import redirect_stderr, redirect_stdout
from .base import BaseCommand, common_aliases, common_flags

PFR_0_9_REQUIREMENTS_FILE_PATH = (
    Path(__file__).parent / "pangeo-forge-recipes-0.9-requirements.txt"
)


class Bake(BaseCommand):
    """
    Command to bake a pangeo forge recipe in a given bakery
    """

    aliases = common_aliases
    flags = common_flags | {
        "prune": (
            {"Bake": {"prune": True}},
            "Prune the recipe to run only for 2 time steps",
        )
    }

    prune = Bool(
        False,
        config=True,
        help="""
        Prune the recipe to only run for 2 time steps.

        Makes it much easier to test recipes!
        """,
    )

    bakery_class = Type(
        default_value=LocalDirectBakery,
        klass=Bakery,
        config=True,
        help="""
        The Bakery to bake this recipe in.

        The Bakery (and its configuration) determine which Apache Beam
        Runner is used, and how options for it are specified.
        Defaults to LocalDirectBakery, which bakes the recipe using Apache
        Beam's "DirectRunner". It doesn't use Docker or the cloud, and runs
        everything locally. Useful only for testing!
        """,
    )

    recipe_id = Unicode(
        None,
        allow_none=True,
        config=True,
        help="""
        Optionally pass this value to run only this recipe_id from the feedstock.

        If empty, all recipes from the feedstock will be run.
        """,
    )

    job_name = Unicode(
        None,
        allow_none=True,
        config=True,
        help="""
        Optionally pass a custom job name for the job run.

        If `None` (the default), a unique name will be generated for the job.
        """,
    )

    @validate("job_name")
    def _validate_job_name(self, proposal):
        """
        Validate that job_name conforms to ^[a-z][-_0-9a-z]{0,62}$ regex.

        That's what is valid in dataflow job names, so let's keep everyone
        in that range.

        Dataflow job names adhere to the following GCP cloud labels requirements:
        https://cloud.google.com/resource-manager/docs/creating-managing-labels#requirements
        """
        validating_regex = r"^[a-z][-_0-9a-z]{0,62}$"
        if not re.match(validating_regex, proposal.value):
            raise ValueError(
                f"job_name must match the regex {validating_regex}, instead found {proposal.value}"
            )
        return proposal.value

    container_image = Unicode(
        "",
        config=True,
        help="""
        Container image to use for this job.

        For GCP DataFlow leaving it blank defaults to letting beam 
        automatically figure out the image to use for the workers 
        based on version of beam and python in use.
        
        For Flink it's required that you pass an beam image
        for the version of python and beam you are targeting
        for example: apache/beam_python3.10_sdk:2.51.0
        more info: https://hub.docker.com/layers/apache/

        Note that some runners (like the local one) may not support this!
        """,
    )

    @validate("container_image")
    def _validate_container_image(self, proposal):
        if self.bakery_class == FlinkOperatorBakery and not proposal.value:
            raise ValueError(
                "'container_name' is required when using the 'FlinkOperatorBakery' "
                "for the version of python and apache-beam you are targeting. "
                "See the sdk images available: https://hub.docker.com/layers/apache/"
            )
        return proposal.value

    def autogenerate_job_name(self):
        """
        Autogenerate a readable job_name
        """
        # special case local checkouts, as no contentprovider is used
        safe_chars = string.ascii_lowercase + string.digits
        if os.path.exists(self.repo):
            name = "local-"
            name += escapism.escape(
                os.path.basename(os.path.abspath(self.repo)),
                safe=safe_chars,
                escape_char="-",
            )
            if self.feedstock_subdir != "feedstock":
                name += "-" + escapism.escape(
                    self.feedstock_subdir, safe=safe_chars, escape_char="-"
                )
            return name.lower()

        # special-case github because it is so common
        if self.repo.startswith("https://github.com/"):
            _, user, repo = self.repo.rsplit("/", 2)
            # Get rid of the '.git' at the end, if it exists
            if repo.endswith(".git"):
                repo = repo[:-4]
            job_name = f"gh-{user}-{repo}-{self.picked_content_provider.content_id}"
        else:
            # everything other than github
            job_name = self.repo
            if self.picked_content_provider.content_id is not None:
                job_name += self.picked_content_provider.content_id

        # Always append current ts to job name, to make it unique
        job_name += "-" + str(int(time.time()))

        return job_name

    def start(self):
        """
        Start the baking process
        """
        if not "pangeo-forge-recipes" in [d.metadata["Name"] for d in distributions()]:
            raise ValueError(
                "To use the `bake` command, `pangeo-forge-recipes` must be installed."
            )
        # Create our storage configurations. Traitlets will do its magic, populate these
        # with appropriate config from config file / commandline / defaults.
        target_storage = TargetStorage(parent=self)
        input_cache_storage = InputCacheStorage(parent=self)
        metadata_cache_storage = MetadataCacheStorage(parent=self)

        self.log.info(
            f"Target Storage is {target_storage}\n", extra={"status": "setup"}
        )
        self.log.info(
            f"Input Cache Storage is {input_cache_storage}\n", extra={"status": "setup"}
        )
        self.log.info(
            f"Metadata Cache Storage is {metadata_cache_storage}\n",
            extra={"status": "setup"},
        )

        injection_specs = get_injectionspecs_from_entrypoints()

        with self.fetch() as checkout_dir:
            if not self.job_name:
                self.job_name = self.autogenerate_job_name()

            injection_values = {
                "TARGET_STORAGE": target_storage.get_forge_target(
                    job_name=self.job_name
                ),
            }

            cache_target = input_cache_storage.get_forge_target(job_name=self.job_name)
            if cache_target:
                injection_values |= {"INPUT_CACHE_STORAGE": cache_target}
            print(injection_values)
            print(injection_specs)

            feedstock = Feedstock(
                Path(checkout_dir) / self.feedstock_subdir,
                prune=self.prune,
                callable_args_injections=get_injections(
                    injection_specs, injection_values
                ),
            )

            self.log.info("Parsing recipes...", extra={"status": "running"})
            with redirect_stderr(self.log, {"status": "running"}), redirect_stdout(
                self.log, {"status": "running"}
            ):
                recipes = feedstock.parse_recipes()

            if self.recipe_id:
                if self.recipe_id not in recipes:
                    raise ValueError(f"{self.recipe_id=} not in {list(recipes)}")
                self.log.info(f"Baking only recipe_id='{self.recipe_id}'")
                recipes = {k: r for k, r in recipes.items() if k == self.recipe_id}

            if self.prune:
                # Prune recipes to only run on certain items if we are asked to
                if hasattr(next(iter(recipes.values())), "copy_pruned"):
                    # pangeo-forge-recipes version < 0.10 has a `copy_pruned` method
                    recipes = {k: r.copy_pruned() for k, r in recipes.items()}

            bakery: Bakery = self.bakery_class(parent=self)

            extra_options = {}

            for name, recipe in recipes.items():
                if len(recipes) > 1:
                    recipe_name_hash = hashlib.sha256(name.encode()).hexdigest()[:5]
                    per_recipe_unique_job_name = (
                        self.job_name[: 62 - 6] + "-" + recipe_name_hash
                    )
                    self.log.info(
                        f"Deploying > 1 recipe. Modifying base {self.job_name = } for recipe "
                        f"{name = } with {recipe_name_hash = }. Submitting job with modified "
                        f"{per_recipe_unique_job_name = }. Note: job names must be <= 63 chars. "
                        "If job_name was > 57 chars, it was truncated to accomodate modification."
                    )
                else:
                    per_recipe_unique_job_name = None

                # if pangeo-forge-recipes is <=0.9, we have to specify a requirements.txt
                # file even if it isn't present, as the image used otherwise will not have pangeo-forge-recipes
                if isinstance(recipe, PTransform):
                    requirements_path = feedstock.feedstock_dir / "requirements.txt"
                    if requirements_path.exists():
                        extra_options["requirements_file"] = str(requirements_path)
                else:
                    extra_options["requirements_file"] = str(
                        PFR_0_9_REQUIREMENTS_FILE_PATH
                    )

                pipeline_options = bakery.get_pipeline_options(
                    job_name=(per_recipe_unique_job_name or self.job_name),
                    # FIXME: Bring this in from meta.yaml?
                    container_image=self.container_image,
                    extra_options=extra_options,
                )

                # Set argv explicitly to empty so Apache Beam doesn't try to parse the commandline
                # for pipeline options - we have traitlets doing that for us.
                pipeline = Pipeline(options=pipeline_options, argv=[])
                # Chain our recipe to the pipeline. This mutates the `pipeline` object!
                # We expect `recipe` to either be a beam PTransform, or an object with a 'to_beam'
                # method that returns a transform.
                if isinstance(recipe, PTransform):
                    # This means we are in pangeo-forge-recipes >=0.9
                    pipeline | recipe
                elif hasattr(recipe, "to_beam"):
                    # We are in pangeo-forge-recipes <=0.9
                    # The import has to be here, as this import is not valid in pangeo-forge-recipes>=0.9
                    # NOTE: `StorageConfig` only requires a target; input and metadata caches are optional,
                    # so those are handled conditionally if provided.
                    from pangeo_forge_recipes.storage import StorageConfig

                    recipe.storage_config = StorageConfig(
                        target_storage.get_forge_target(job_name=self.job_name),
                    )
                    for attrname, optional_storage in zip(
                        ("cache", "metadata"),
                        (input_cache_storage, metadata_cache_storage),
                    ):
                        # `.root_path` is an empty string by default, so if the user has not setup this
                        # optional storage type in config, this block is skipped.
                        if optional_storage.root_path:
                            setattr(
                                recipe.storage_config,
                                attrname,
                                optional_storage.get_forge_target(
                                    job_name=self.job_name
                                ),
                            )
                    # with configured storage now attached, compile recipe to beam
                    pipeline | recipe.to_beam()

                # Some bakeries are blocking - if Beam is configured to use them, calling
                # pipeline.run() blocks. Some are not. We handle that here, and provide
                # appropriate feedback to the user too.
                extra = {
                    "recipe": name,
                    "job_name": (per_recipe_unique_job_name or self.job_name),
                }
                if bakery.blocking:
                    self.log.info(
                        f"Running job for recipe {name}\n",
                        extra=extra | {"status": "running"},
                    )
                    pipeline.run()
                else:
                    result = pipeline.run()
                    job_id = result.job_id()
                    self.log.info(
                        f"Submitted job {job_id} for recipe {name}",
                        extra=extra | {"job_id": job_id, "status": "submitted"},
                    )



# Validate rewriting works in the presence of multiple aliased imports
import apache_beam as random_name_because_why_not
import apache_beam as random_name_because_why_not_second_fucking_time
from pangeo_forge_recipes.patterns import pattern_from_file_sequence

pattern = pattern_from_file_sequence(["hi.nc"])

recipe = random_name_because_why_not.Create(pattern.items())
recipe = random_name_because_why_not_second_fucking_time.Create(pattern.items())



import json
import subprocess


def test_json_error():
    """
    Test that app exceptions still show up as JSON
    """
    # Do soemthing that causes an exception, and test we get that info out
    cmd = [
        "pangeo-forge-runner",
        "expand-meta",
        "--repo",
        "https://example.com",  # This isn't a valid repo
        "--json",
    ]
    proc = subprocess.run(cmd, encoding="utf-8", capture_output=True)
    assert proc.returncode == 1
    found_exception = False
    for l in proc.stdout.splitlines():
        p = json.loads(l)
        if p["status"] == "failed":
            assert "exc_info" in p
            found_exception = True
    assert found_exception



import json
import os
import secrets
import subprocess

invocations = [
    {
        "repo": "https://github.com/pforgetest/gpcp-from-gcs-feedstock.git",
        "ref": "4f41e02512b2078c8bdb286368a1a9d878b5cec2",
        "meta": {
            "title": "Global Precipitation Climatology Project",
            "description": "Global Precipitation Climatology Project (GPCP) Daily Version 1.3 gridded, merged ty satellite/gauge precipitation Climate data Record (CDR) from 1996 to present.\n",
            "recipes": [{"id": "gpcp-from-gcs", "object": "recipe:recipe"}],
            "provenance": {
                "providers": [
                    {
                        "name": "NOAA NCEI",
                        "description": "National Oceanographic & Atmospheric Administration National Centers for Environmental Information",
                        "roles": ["host", "licensor"],
                        "url": "https://www.ncei.noaa.gov/products/global-precipitation-climatology-project",
                    },
                    {
                        "name": "University of Maryland",
                        "description": "University of Maryland College Park Earth System Science Interdisciplinary Center (ESSIC) and Cooperative Institute for Climate and Satellites (CICS).\n",
                        "roles": ["producer"],
                        "url": "http://gpcp.umd.edu/",
                    },
                ],
                "license": "No constraints on data access or use.",
            },
            "maintainers": [
                {
                    "name": "Charles Stern",
                    "orcid": "0000-0002-4078-0852",
                    "github": "cisaacstern",
                }
            ],
        },
    }
]


def test_expand_meta_json():
    for invocation in invocations:
        cmd = [
            "pangeo-forge-runner",
            "expand-meta",
            "--repo",
            invocation["repo"],
            "--ref",
            invocation["ref"],
            "--json",
        ]
        out = subprocess.check_output(cmd, encoding="utf-8")
        found_meta = False
        for l in out.splitlines():
            p = json.loads(l)
            if p["status"] == "completed":
                assert p["meta"] == invocation["meta"]
                found_meta = True
        assert found_meta


def test_expand_meta_no_json():
    for invocation in invocations:
        cmd = [
            "pangeo-forge-runner",
            "expand-meta",
            "--repo",
            invocation["repo"],
            "--ref",
            invocation["ref"],
        ]
        out = subprocess.check_output(cmd, encoding="utf-8")
        last_line = out.splitlines()[-1]
        assert json.loads(last_line) == invocation["meta"]


def test_missing_config_file():
    non_existent_path = secrets.token_hex() + ".py"
    assert not os.path.exists(non_existent_path)
    cmd = ["pangeo-forge-runner", "expand-meta", "--config", non_existent_path]
    proc = subprocess.run(cmd, encoding="utf-8", capture_output=True, text=True)
    assert proc.returncode == 1
    assert "Could not read config from file" in proc.stderr



from unittest.mock import patch

from pangeo_forge_runner.bakery.flink import FlinkOperatorBakery


def test_pipelineoptions():
    """
    Quickly validate some of the PipelineOptions set
    """
    fob = FlinkOperatorBakery()
    fob.parallelism = 100
    fob.max_parallelism = 100

    # FlinkOperatorBakery.get_pipeline_options calls `kubectl` in a subprocess,
    # so we patch subprocess here to skip that behavior for this test
    with patch("pangeo_forge_runner.bakery.flink.subprocess"):
        po = fob.get_pipeline_options("job", "some-container:some-tag", {})
        # some flink args, e.g. 'parallelism', are apparently 'unknown_options' from
        # the perspective of PipelineOptions, so we retain those here for the test.
        # it doesn't seem like their 'unknown' status prevents them from being passed to
        # flink in an actual deployment, though.
        opts = po.get_all_options(retain_unknown_options=True)

    assert opts["parallelism"] == 100
    assert opts["max_parallelism"] == 100



import hashlib
import json
import re
import subprocess
import sys
import tempfile
from importlib.metadata import distributions, version

import pytest
import xarray as xr
from packaging.version import parse as parse_version

from pangeo_forge_runner.commands.bake import Bake


@pytest.fixture
def recipes_uninstalled():
    """Uninstall `pangeo-forge-recipes` for `test_bake_requires_recipes_installed`."""
    # first confirm that it's installed to begin with
    assert "pangeo-forge-recipes" in [d.metadata["Name"] for d in distributions()]
    # and capture the version, which we'll reinstall after the test
    recipes_version = parse_version(version("pangeo-forge-recipes"))
    # now uninstall it
    uninstall = subprocess.run(
        f"{sys.executable} -m pip uninstall pangeo-forge-recipes -y".split()
    )
    assert uninstall.returncode == 0
    assert "pangeo-forge-recipes" not in [d.metadata["Name"] for d in distributions()]
    # and yield to the test
    yield True
    # test is complete, now reinstall pangeo-forge-recipes in the test env
    reinstall = subprocess.run(
        f"{sys.executable} -m pip install pangeo-forge-recipes=={recipes_version}".split()
    )
    assert reinstall.returncode == 0
    # make sure it's there, and in the expected version
    assert "pangeo-forge-recipes" in [d.metadata["Name"] for d in distributions()]
    assert parse_version(version("pangeo-forge-recipes")) == recipes_version


def test_bake_requires_recipes_installed(recipes_uninstalled):
    """`pangeo-forge-runner` does not require `pangeo-forge-recipes` to be installed,
    but `pangeo-forge-recipes` *is* required to use the `bake` command, so test that
    we get a descriptive error if we try to invoke this command without it installed.
    """
    assert recipes_uninstalled
    bake = Bake()
    with pytest.raises(
        ValueError,
        match="To use the `bake` command, `pangeo-forge-recipes` must be installed.",
    ):
        bake.start()


@pytest.mark.parametrize(
    "job_name, raises",
    (
        ["valid-job", False],
        ["valid_job", False],
        ["".join(["a" for i in range(63)]), False],  # <= 63 chars allowed
        ["".join(["a" for i in range(64)]), True],  # > 63 chars not allowed
        ["invali/d", True],  # dashes are the only allowable punctuation
        ["1valid-job", True],  # can only start with letters
        ["-valid-job", True],  # can only start with letters
        ["Valid-Job", True],  # uppercase letters not allowed
    ),
)
def test_job_name_validation(job_name, raises):
    bake = Bake()
    if raises:
        with pytest.raises(
            ValueError,
            match=re.escape(
                f"job_name must match the regex ^[a-z][-_0-9a-z]{{0,62}}$, instead found {job_name}"
            ),
        ):
            bake.job_name = job_name
    else:
        bake.job_name = job_name
        assert bake.job_name == job_name


@pytest.mark.parametrize(
    "container_image, raises",
    (
        ["", True],
        ["apache/beam_python3.10_sdk:2.51.0", False],
    ),
)
def test_container_name_validation(container_image, raises):
    bake = Bake()
    if raises:
        with pytest.raises(
            ValueError,
            match=r"^'container_name' is required.*",
        ):
            bake.bakery_class = "pangeo_forge_runner.bakery.flink.FlinkOperatorBakery"
            bake.container_image = container_image
    else:
        bake.bakery_class = "pangeo_forge_runner.bakery.flink.FlinkOperatorBakery"
        bake.container_image = container_image
        assert bake.container_image == container_image


@pytest.fixture(params=["recipe_object", "dict_object"])
def recipes_version_ref(request):
    pfr_version = parse_version(version("pangeo-forge-recipes"))
    if pfr_version >= parse_version("0.10"):
        recipes_version_ref = "0.10.x"
    else:
        recipes_version_ref = "0.9.x"
    return (
        recipes_version_ref
        if not request.param == "dict_object"
        else f"{recipes_version_ref}-dictobj"
    )


@pytest.mark.parametrize(
    ("recipe_id", "expected_error", "custom_job_name"),
    (
        [None, None, None],
        ["gpcp-from-gcs", None, None],
        [
            "invalid_recipe_id",
            "ValueError: self.recipe_id='invalid_recipe_id' not in ['gpcp-from-gcs']",
            None,
        ],
        [None, None, "special-name-for-job"],
    ),
)
def test_gpcp_bake(
    minio, recipe_id, expected_error, custom_job_name, recipes_version_ref
):
    if recipes_version_ref == "0.9.x-dictobj" or (
        recipes_version_ref == "0.10.x-dictobj" and recipe_id
    ):
        # TODO: clarify fixturing story to avoid this hackiness
        pytest.skip(
            "We only test dictobjs for recipes >0.10.0, and without recipe_id's"
        )

    fsspec_args = {
        "key": minio["username"],
        "secret": minio["password"],
        "client_kwargs": {"endpoint_url": minio["endpoint"]},
    }

    config = {
        "Bake": {
            "prune": True,
            "bakery_class": "pangeo_forge_runner.bakery.local.LocalDirectBakery",
        },
        "TargetStorage": {
            "fsspec_class": "s3fs.S3FileSystem",
            "fsspec_args": fsspec_args,
            "root_path": "s3://gpcp/target/",
        },
        "InputCacheStorage": {
            "fsspec_class": "s3fs.S3FileSystem",
            "fsspec_args": fsspec_args,
            "root_path": "s3://gpcp/input-cache/",
        },
        "MetadataCacheStorage": {
            "fsspec_class": "s3fs.S3FileSystem",
            "fsspec_args": fsspec_args,
            "root_path": "s3://gpcp/metadata-cache/",
        },
    }

    if recipe_id:
        config["Bake"].update({"recipe_id": recipe_id})
    if custom_job_name:
        config["Bake"].update({"job_name": custom_job_name})

    with tempfile.NamedTemporaryFile("w", suffix=".json") as f:
        json.dump(config, f)
        f.flush()
        cmd = [
            "pangeo-forge-runner",
            "bake",
            "--repo",
            "https://github.com/pforgetest/gpcp-from-gcs-feedstock.git",
            "--ref",
            # in the test feedstock, tags are named for the recipes version
            # which was used to write the recipe module
            recipes_version_ref,
            "--json",
            "-f",
            f.name,
        ]
        proc = subprocess.run(cmd, capture_output=True)
        stdout = proc.stdout.decode().splitlines()

        if expected_error:
            assert proc.returncode == 1
            stdout[-1] == expected_error

        else:
            assert proc.returncode == 0

            job_name_logs = [
                json.loads(line) for line in stdout if "Running job for recipe " in line
            ]
            job_names = {line["recipe"]: line["job_name"] for line in job_name_logs}
            for recipe_name, job_name in job_names.items():
                if custom_job_name:
                    assert job_name.startswith(custom_job_name)
                else:
                    assert job_name.startswith("gh-pforgetest-gpcp-from-gcs-")

                if "dictobj" in recipes_version_ref:
                    assert job_name.endswith(
                        hashlib.sha256(recipe_name.encode()).hexdigest()[:5]
                    )

            # In pangeo-forge-recipes>=0.10.0, the actual zarr store is produced in a
            # *subpath* of target_storage.rootpath, rather than in the
            # root path itself. This is a compatibility break vs the previous
            # versions of pangeo-forge-recipes. https://github.com/pangeo-forge/pangeo-forge-recipes/pull/495
            # has more information

            if recipes_version_ref == "0.10.x":
                zarr_store_full_paths = [config["TargetStorage"]["root_path"] + "gpcp/"]
            elif recipes_version_ref == "0.10.x-dictobj":
                zarr_store_root_path = config["TargetStorage"]["root_path"]
                zarr_store_full_paths = [
                    zarr_store_root_path + store_name
                    for store_name in ["gpcp-dict-key-0", "gpcp-dict-key-1"]
                ]
            else:
                zarr_store_full_paths = [config["TargetStorage"]["root_path"]]

            # Open the generated datasets with xarray!
            for path in zarr_store_full_paths:
                print(f"Opening dataset for {path = }")
                ds = xr.open_dataset(
                    # We specify a store_name of "gpcp" in the test recipe
                    path,
                    backend_kwargs={"storage_options": fsspec_args},
                    engine="zarr",
                )

                assert (
                    ds.title
                    == "Global Precipitation Climatatology Project (GPCP) Climate Data Record (CDR), Daily V1.3"
                )
                # --prune prunes to two time steps by default, so we expect 2 items here
                assert len(ds.precip) == 2
                print(ds)

            # `mc` isn't the best way, but we want to display all the files in our minio
            with tempfile.TemporaryDirectory() as mcd:
                cmd = [
                    "mc",
                    "--config-dir",
                    mcd,
                    "alias",
                    "set",
                    "local",
                    minio["endpoint"],
                    minio["username"],
                    minio["password"],
                ]

                subprocess.run(cmd, check=True)

                cmd = ["mc", "--config-dir", mcd, "ls", "--recursive", "local"]
                subprocess.run(cmd, check=True)



from textwrap import dedent

import pytest
from ruamel.yaml import YAML

from pangeo_forge_runner.feedstock import Feedstock
from pangeo_forge_runner.meta_yaml import MetaYaml

yaml = YAML()


@pytest.fixture(params=["recipe_object", "dict_object", "both"])
def tmp_feedstock(request, tmp_path_factory: pytest.TempPathFactory):
    tmpdir = tmp_path_factory.mktemp("feedstock")
    if request.param == "recipe_object":
        meta_yaml = dedent(
            """\
        recipes:
          - id: aws-noaa-sea-surface-temp-whoi
            object: 'recipe:recipe'
        """
        )
        recipe_py = dedent(
            """\
        class Recipe:
          pass

        recipe = Recipe()
        """
        )
    elif request.param == "dict_object":
        meta_yaml = dedent(
            """\
        recipes:
          - dict_object: 'recipe:recipes'
        """
        )
        recipe_py = dedent(
            """\
        class Recipe:
          pass

        recipes = {"my_recipe": Recipe()}
        """
        )

    elif request.param == "both":
        meta_yaml = dedent(
            """\
        recipes:
          - id: aws-noaa-sea-surface-temp-whoi
            object: 'recipe:recipe'
          - dict_object: 'recipe:recipes'
        """
        )
        recipe_py = dedent(
            """\
        class Recipe:
          pass

        recipe = Recipe()
        recipes = {"my_recipe": Recipe()}
        """
        )

    with open(tmpdir / "meta.yaml", mode="w") as f:
        f.write(meta_yaml)
    with open(tmpdir / "recipe.py", mode="w") as f:
        f.write(recipe_py)

    yield tmpdir, meta_yaml, request.param


def test_feedstock(tmp_feedstock):
    tmpdir, meta_yaml, recipes_section_type = tmp_feedstock
    f = Feedstock(feedstock_dir=tmpdir)
    # equality of HasTraits instances doesn't work as I might expect,
    # so just check equality of the relevant trait (`.recipes`)
    assert f.meta.recipes == MetaYaml(**yaml.load(meta_yaml)).recipes

    recipes = f.parse_recipes()

    # the recipe_object metadata  looks something like this:
    #   {'recipes': [{'id': 'my_recipe', 'object': 'DICT_VALUE_PLACEHOLDER'}]}
    # and the dict_object metadata looks like this:
    #   {'recipes': [{'id': 'aws-noaa-sea-surface-temp-whoi', 'object': 'recipe:recipe'}]}
    if recipes_section_type == "recipe_object":
        expanded_meta = f.get_expanded_meta()
        assert expanded_meta["recipes"] == [
            {"id": "aws-noaa-sea-surface-temp-whoi", "object": "recipe:recipe"},
        ]
    elif recipes_section_type == "dict_object":
        expanded_meta = f.get_expanded_meta()
        assert expanded_meta["recipes"] == [
            {"id": "my_recipe", "object": "DICT_VALUE_PLACEHOLDER"},
        ]
    elif recipes_section_type == "both":
        with pytest.raises(NotImplementedError):
            _ = f.get_expanded_meta()

    for r in recipes.values():
        # the values of the recipes dict should all be python objects
        # we used the mock type `Recipe` here, so this should be true:
        assert str(r).startswith("<Recipe object")



import ast
from pathlib import Path

import pytest

from pangeo_forge_runner.recipe_rewriter import RecipeRewriter

HERE = Path(__file__).parent


@pytest.mark.parametrize(
    ("name", "params_path", "original_path", "rewritten_path"),
    [
        (
            case.name,
            HERE / f"rewriter-tests/{case.name}/params.py",
            HERE / f"rewriter-tests/{case.name}/original.py",
            HERE / f"rewriter-tests/{case.name}/rewritten.py",
        )
        for case in (HERE / "rewriter-tests").iterdir()
        if case.is_dir() and not case.name.startswith("__")
    ],
)
def test_rewriter(name, params_path: Path, original_path: Path, rewritten_path: Path):
    params = {}

    exec(compile(params_path.read_text(), filename=params_path, mode="exec"), params)

    orig_ast = ast.parse(
        source=original_path.read_text(), filename=original_path, mode="exec"
    )
    rewritten_ast = ast.parse(
        source=rewritten_path.read_text(), filename=rewritten_path, mode="exec"
    )

    rewriter = RecipeRewriter(**params["params"])

    assert ast.unparse(rewriter.visit(orig_ast)) == ast.unparse(rewritten_ast)



import os
import tempfile

import pytest
from repo2docker import contentproviders

from pangeo_forge_runner.commands.base import BaseCommand


def test_bad_cp():
    """
    Test we throw a useful exception if we can't get a ContentProvider

    This doesn't happen with the default configuration as we assume
    any URL is actually a git URL if we can't prove otherwise
    """
    bc = BaseCommand()
    bc.content_providers.remove(contentproviders.Git)

    bc.repo = "https://example.com"
    with pytest.raises(ValueError):
        with bc.fetch() as _:
            pass


def test_local():
    """
    Test that local directories are just treated as nops
    """
    with tempfile.TemporaryDirectory() as d:
        bc = BaseCommand()
        bc.repo = d

        with bc.fetch() as checkout_dir:
            assert checkout_dir == d

        # Make sure the directory isn't deleted either by the end!
        assert os.path.exists(d)



def some_callable(some_argument):
    pass


recipes = {
    "a": some_callable(
        some_argument=_CALLABLE_ARGS_INJECTIONS.get("some_callable", {}).get(  # noqa
            "some_argument"
        )
    ),
    "b": some_callable(
        some_argument=_CALLABLE_ARGS_INJECTIONS.get("some_callable", {}).get(  # noqa
            "some_argument"
        )
    ),
}



import json
import subprocess
import tempfile
import time
from importlib.metadata import version

import pytest
import xarray as xr
from packaging.version import parse as parse_version


def test_dataflow_integration():
    pfr_version = parse_version(version("pangeo-forge-recipes"))
    if pfr_version >= parse_version("0.10"):
        recipe_version_ref = str(pfr_version)
    else:
        recipe_version_ref = "0.9.x"
    bucket = "gs://pangeo-forge-runner-ci-testing"
    config = {
        "Bake": {
            "prune": True,
            "bakery_class": "pangeo_forge_runner.bakery.dataflow.DataflowBakery",
        },
        "DataflowBakery": {"temp_gcs_location": bucket + "/temp"},
        "TargetStorage": {
            "fsspec_class": "gcsfs.GCSFileSystem",
            "root_path": bucket + "/target/{job_name}",
        },
        "InputCacheStorage": {
            "fsspec_class": "gcsfs.GCSFileSystem",
            "root_path": bucket + "/input-cache/{job_name}",
        },
        "MetadataCacheStorage": {
            "fsspec_class": "gcsfs.GCSFileSystem",
            "root_path": bucket + "/metadata-cache/{job_name}",
        },
    }

    with tempfile.NamedTemporaryFile("w", suffix=".json") as f:
        json.dump(config, f)
        f.flush()
        cmd = [
            "pangeo-forge-runner",
            "bake",
            "--repo",
            "https://github.com/pforgetest/gpcp-from-gcs-feedstock.git",
            "--ref",
            # in the test feedstock, tags are named for the recipes version
            # which was used to write the recipe module
            recipe_version_ref,
            "--json",
            "-f",
            f.name,
        ]
        print("\nSubmitting job...")
        submit_proc = subprocess.run(cmd, capture_output=True)
        assert submit_proc.returncode == 0
        lastline = json.loads(submit_proc.stdout.decode().splitlines()[-1])
        assert lastline["status"] == "submitted"
        job_id = lastline["job_id"]
        job_name = lastline["job_name"]
        print(f"Job submitted with {job_id = }")
        # note the start time, because certain errors on dataflow manifest as long hangs,
        # and if that's the case, we'll want to bail out of this test manually, rather than
        # wait for the the job to officially fail.
        start = time.time()

        # 6 minutes seems like an average runtime for these jobs, but being optimistic
        # let's start by waiting 5 minutes
        print(f"Waiting for 5 mins, starting at {start = }")
        time.sleep(60 * 5)

        # okay, time to start checking if the job is done
        show_job = f"gcloud dataflow jobs show {job_id} --format=json".split()
        while True:
            elapsed = time.time() - start
            print(f"Time {elapsed = }")
            if elapsed > 60 * 12:
                # if 12 minutes have elapsed (twice the expected time to complete the job),
                # we're going to assume the job is hanging, and call this test a failure.
                # remember: we're sourcing data for this job from within GCS, so networking
                # shouldn't delay things *too* much. if we eventually find that jobs may take
                # more than 12 minutes and not be hanging, we can change this assumption.
                pytest.fail(f"Time {elapsed = } exceedes 12 minutes.")

            # check job state
            state_proc = subprocess.run(show_job, capture_output=True)
            assert state_proc.returncode == 0
            state = json.loads(state_proc.stdout)["state"]
            print(f"Current {state = }")
            if state == "Done":
                # on Dataflow, "Done" means success
                break
            elif state == "Running":
                # still running, let's give it another 30s then check again
                time.sleep(30)
            else:
                # consider any other state a failure
                pytest.fail(f"{state = } is neither 'Done' nor 'Running'")

        # open the generated dataset with xarray!
        target_path = config["TargetStorage"]["root_path"].format(job_name=job_name)
        if pfr_version >= parse_version("0.10"):
            # in pangeo-forge-eecipes>=0.10.0, an additional `StoreToZarr.store_name` kwarg
            # is appended to the formatted root path at execution time. for ref `0.10.x`,
            # the value of that kwarg is "gpcp", so we append that here.
            target_path += "/gpcp"
        gpcp = xr.open_dataset(target_path, engine="zarr")

        assert (
            gpcp.title
            == "Global Precipitation Climatatology Project (GPCP) Climate Data Record (CDR), Daily V1.3"
        )
        # --prune prunes to two time steps by default, so we expect 2 items here
        assert len(gpcp.precip) == 2
        print(gpcp)



"""
AST rewrites for recipe files to support injecting config

FIXME:

Still needs to handle a couple of additional cases.
- `from apache_beam import Create` is not handled for pruning yet. Need
  to handle `ImportFrom` statements.
- `import pangeo_forge_recipes; pangeo_forge_recipes.transforms.StoreToZarr` is not handled either
"""
from ast import (
    Attribute,
    Call,
    Constant,
    Dict,
    Import,
    Load,
    Name,
    NodeTransformer,
    fix_missing_locations,
    keyword,
)
from typing import Optional


class RecipeRewriter(NodeTransformer):
    """
    Transform a recipe file to provide 'configuration' as needed.
    """

    def __init__(
        self, prune: bool = False, callable_args_injections: Optional[dict] = None
    ):
        """
        prune: Set to true to add a .prune() call to FilePatterns passed to beam.Create
        callable_args_injections: A dict of callable names (as keys) with injected kwargs as value
        """
        self.prune = prune
        self.callable_args_injections = (
            callable_args_injections if callable_args_injections else {}
        )

        self._import_aliases = {}

    def visit_Import(self, node: Import) -> Import:
        for name in node.names:
            if name.asname:
                self._import_aliases[name.asname] = name.name
            else:
                self._import_aliases[name.name] = name.name

        return node

    def get_exec_globals(self):
        """
        Return a dict with objects to be injected into recipe while executing.

        Should be passed to `globals` of `exec` function
        """
        # This is used by our transformations to inject parameters to callables
        return {"_CALLABLE_ARGS_INJECTIONS": self.callable_args_injections}

    def transform_prune(self, node: Call) -> Call:
        """
        Transform a FilePattern object being passed to beam.Create to call a .prune() method

        node: A ast.Call object representing the `beam.Create` call
        """
        if not self.prune:
            return node
        # The object on which `.items()` is being called is what we will consider our `FilePattern` object
        # We will transform this .items() call into a .prune().items() call
        file_pattern_obj = node.args[0].func.value

        items_call = Call(
            func=Attribute(
                value=Call(
                    func=Attribute(value=file_pattern_obj, attr="prune", ctx=Load()),
                    args=[],
                    keywords=[],
                ),
                attr="items",
                ctx=Load(),
            ),
            args=[],
            keywords=[],
        )

        node.args = [items_call]

        return node

    def _make_injected_get(
        self, injected_variable: str, callable_name: str, param_name: str
    ) -> Call:
        return Call(
            func=Attribute(
                value=Call(
                    func=Attribute(
                        value=Name(id=injected_variable, ctx=Load()),
                        attr="get",
                        ctx=Load(),
                    ),
                    args=[
                        Constant(value=callable_name),
                        Dict(keys=[], values=[]),
                    ],
                    keywords=[],
                ),
                attr="get",
                ctx=Load(),
            ),
            args=[Constant(value=param_name)],
            keywords=[],
        )

    def visit_Call(self, node: Call) -> Call:
        """
        Rewrite calls that return a FilePattern if we need to prune them
        """
        if isinstance(node.func, Attribute):
            # FIXME: Support it being imported as from apache_beam import Create too
            if "apache_beam" not in self._import_aliases.values():
                # if beam hasn't been imported, don't rewrite anything
                return node

            # Only rewrite parameters to apache_beam.Create, regardless
            # of how it is imported as
            if node.func.attr == "Create" and (
                self._import_aliases.get(node.func.value.id) == "apache_beam"
            ):
                # If there is a single argument pased to beam.Create, and it is <something>.items()
                # This is the heurestic we use for figuring out that we are in fact operating on a FilePattern object
                if (
                    len(node.args) == 1
                    and isinstance(node.args[0].func, Attribute)
                    and node.args[0].func.attr == "items"
                ):
                    return fix_missing_locations(self.transform_prune(node))
        elif isinstance(node.func, Name):
            # FIXME: Support importing in other ways
            for name, params in self.callable_args_injections.items():
                if name == node.func.id:
                    node.keywords += [
                        keyword(
                            arg=k,
                            value=self._make_injected_get(
                                "_CALLABLE_ARGS_INJECTIONS", name, k
                            ),
                        )
                        for k in params
                    ]
            return fix_missing_locations(node)

        return node



import pytest

from pangeo_forge_runner.bakery.base import Bakery


def test_unimplemented():
    b = Bakery()
    with pytest.raises(NotImplementedError):
        b.get_pipeline_options("test", "test", {})



import copy
from textwrap import dedent

import pytest
from ruamel.yaml import YAML
from traitlets import TraitError

from pangeo_forge_runner.meta_yaml import MetaYaml

yaml = YAML()


@pytest.fixture
def with_recipes_list() -> str:
    return dedent(
        """\
    title: 'AWS NOAA WHOI SST'
    description: 'Analysis-ready datasets derived from AWS NOAA WHOI NetCDF'
    recipes:
      - id: aws-noaa-sea-surface-temp-whoi
        object: 'recipe:recipe'
    provenance:
      providers:
        - name: 'AWS NOAA Oceanic CDR'
          description: 'Registry of Open Data on AWS National Oceanographic & Atmospheric Administration National Centers for Environmental Information'
          roles:
            - producer
            - licensor
          url: s3://noaa-cdr-sea-surface-temp-whoi-pds/
      license: 'Open Data'
    maintainers:
      - name: 'Jo Contributor'
        orcid: '0000-0000-0000-0000'
        github: jocontributor123
    """  # noqa: E501
    )


@pytest.fixture
def valid_meta_yaml(with_recipes_list: str) -> dict:
    return yaml.load(with_recipes_list)


@pytest.fixture
def valid_meta_yaml_dict_object(with_recipes_list: str) -> dict:
    with_dict_object = with_recipes_list.replace(
        dedent(
            """\
        recipes:
          - id: aws-noaa-sea-surface-temp-whoi
            object: 'recipe:recipe'
        """
        ),
        dedent(
            """\
        recipes:
          - dict_object: 'recipe:recipes'
        """
        ),
    )
    return yaml.load(with_dict_object)


def test_schema_valid(valid_meta_yaml):
    _ = MetaYaml(**valid_meta_yaml)


def test_schema_valid_dict_object(valid_meta_yaml_dict_object):
    _ = MetaYaml(**valid_meta_yaml_dict_object)


@pytest.mark.parametrize(
    "field",
    [
        "title",
        "description",
        "recipes",
        "provenance",
        "maintainers",
    ],
)
def test_missing_toplevel_field(valid_meta_yaml, field):
    meta_yaml_copy = copy.deepcopy(valid_meta_yaml)
    del meta_yaml_copy[field]
    if field == "recipes":
        # ``recipes`` is the only required field
        with pytest.raises(TraitError):
            _ = MetaYaml(**meta_yaml_copy)
    else:
        # all others fields can be left empty without raising an error
        _ = MetaYaml(**meta_yaml_copy)


@pytest.mark.parametrize(
    "subfield",
    [
        "id",
        "object",
    ],
)
def test_missing_recipes_subfield(valid_meta_yaml, subfield):
    invalid_meta_yaml = copy.deepcopy(valid_meta_yaml)
    del invalid_meta_yaml["recipes"][0][subfield]

    with pytest.raises(TraitError):
        _ = MetaYaml(**invalid_meta_yaml)


def test_recipes_field_cannot_be_empty_container():
    with pytest.raises(TraitError):
        _ = MetaYaml(recipes=[])


def test_recipes_field_invalid_keys():
    with pytest.raises(TraitError):
        # "dict_object" key can't be used in combination with other keys
        _ = MetaYaml(recipes={"id": "abcdefg", "dict_object": "abc:def"})
    with pytest.raises(TraitError):
        # the only valid keys are {"id", "object"} together,
        # or "dict_object" alone. other keys are not allowed.
        _ = MetaYaml(recipes={"some_other_key": "abc:def"})


# TODO: In a future "strict" mode, ensure provenance fields are all provided.
# --------------------------------------------------------------------------
# @pytest.mark.parametrize(
#     "subfield",
#     [
#         "providers",
#         "license",
#     ],
# )
# def test_missing_provenance_subfield(valid_meta_yaml, subfield, schema):
#     invalid_meta_yaml = copy.deepcopy(valid_meta_yaml)
#     del invalid_meta_yaml["provenance"][subfield]
#
#     with pytest.raises(TraitError):
#         _ = MetaYaml(**meta_yaml_copy)


# TODO: In a future "strict" mode, ensure providers fields are all provided.
# --------------------------------------------------------------------------
# @pytest.mark.parametrize(
#     "subfield",
#     [
#         "name",
#         "description",
#         "roles",
#         "url",
#     ],
# )
# def test_missing_providers_subfield(valid_meta_yaml, subfield, schema):
#     invalid_meta_yaml = copy.deepcopy(valid_meta_yaml)
#     del invalid_meta_yaml["provenance"]["providers"][0][subfield]
#
#     with pytest.raises(ValidationError, match=f"'{subfield}' is a required property"):
#         _ = MetaYaml(**meta_yaml_copy)


# TODO: In a future "strict" mode, ensure maintainers fields are all provided.
# ----------------------------------------------------------------------------
# @pytest.mark.parametrize(
#     "subfield",
#     [
#         "name",
#         "orcid",
#         "github",
#     ],
# )
# def test_missing_maintainers_subfield(valid_meta_yaml, subfield, schema):
#     invalid_meta_yaml = copy.deepcopy(valid_meta_yaml)
#     del invalid_meta_yaml["maintainers"][0][subfield]
#
#     with pytest.raises(TraitError):
#         _ = MetaYaml(**meta_yaml_copy)



import json
import subprocess
import tempfile
import time
from importlib.metadata import version

import pytest
import xarray as xr
from packaging.version import parse as parse_version


def test_flink_bake(minio_service, flinkversion, pythonversion, beamversion):
    fsspec_args = {
        "key": minio_service["username"],
        "secret": minio_service["password"],
        "client_kwargs": {"endpoint_url": minio_service["endpoint"]},
    }

    pfr_version = parse_version(version("pangeo-forge-recipes"))
    if pfr_version >= parse_version("0.10"):
        recipe_version_ref = str(pfr_version)
    else:
        recipe_version_ref = "0.9.x"
        pytest.xfail(
            f"{pfr_version = }, which is < 0.10. "
            "Flink tests timeout with this recipes version, so we xfail this test."
        )

    bucket = "s3://gpcp-out"
    config = {
        "Bake": {
            "prune": True,
            "job_name": "recipe",
            "bakery_class": "pangeo_forge_runner.bakery.flink.FlinkOperatorBakery",
            # there must be a job-server jar available for the matching
            # `apache-beam` and `FlinkOperatorBakery.flink_version` here:
            # https://repo.maven.apache.org/maven2/org/apache/beam/beam-runners-flink-1.16-job-server/
            "container_image": f"apache/beam_python{pythonversion}_sdk:{beamversion}",
        },
        "TargetStorage": {
            "fsspec_class": "s3fs.S3FileSystem",
            "fsspec_args": fsspec_args,
            "root_path": bucket + "/target/{job_name}",
        },
        "InputCacheStorage": {
            "fsspec_class": "s3fs.S3FileSystem",
            "fsspec_args": fsspec_args,
            "root_path": bucket + "/input-cache/{job_name}",
        },
        "MetadataCacheStorage": {
            "fsspec_class": "s3fs.S3FileSystem",
            "fsspec_args": fsspec_args,
            "root_path": bucket + "/metadata-cache/{job_name}",
        },
        "FlinkOperatorBakery": {
            "flink_version": flinkversion,
            "job_manager_resources": {"memory": "1024m", "cpu": 0.30},
            "task_manager_resources": {"memory": "2048m", "cpu": 0.30},
            "parallelism": 1,
            "flink_configuration": {
                "taskmanager.numberOfTaskSlots": "1",
                "taskmanager.memory.jvm-overhead.max": "2048m",
            },
        },
    }

    with tempfile.NamedTemporaryFile("w", suffix=".json") as f:
        json.dump(config, f)
        f.flush()
        cmd = [
            "pangeo-forge-runner",
            "bake",
            "--repo",
            "https://github.com/pforgetest/gpcp-from-gcs-feedstock.git",
            "--ref",
            # in the test feedstock, tags are named for
            # the recipe version used to write the recipe module
            recipe_version_ref,
            "-f",
            f.name,
        ]

        print("\nSubmitting job...")
        timeout = 60 * 4
        with subprocess.Popen(
            cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE, text=True
        ) as proc:
            start = time.time()
            for line in proc.stdout:
                # nice to have output
                print(line, end="")

                elapsed_time = time.time() - start
                if elapsed_time >= timeout:
                    raise Exception("timeout reached, exiting")

        # make sure the last time submitted job
        assert line.startswith("Started Flink job as")

        # use minio cli to inuit when job is finished after a waiting period
        # TODO: we need to get the historyserver up so we can query job status async
        # https://nightlies.apache.org/flink/flink-docs-release-1.16/docs/deployment/advanced/historyserver/
        time.sleep(60 * 2)
        cmd = [
            "mc",
            "alias",
            "set",
            "myminio",
            minio_service["endpoint"],
            minio_service["username"],
            minio_service["password"],
        ]
        proc = subprocess.run(cmd, capture_output=True)
        assert proc.returncode == 0

        # set up path lookups for minio cli and xarray
        target_path = config["TargetStorage"]["root_path"].format(
            job_name=config["Bake"]["job_name"]
        )
        if pfr_version >= parse_version("0.10"):
            # in pangeo-forge-recipes>=0.10.0, an additional `StoreToZarr.store_name` kwarg
            # is appended to the formatted root path at execution time. for ref `0.10.x`,
            # the value of that kwarg is "gpcp", so we append that here.
            target_path += "/gpcp"

        cmd = [
            "mc",
            "ls",
            "myminio/{}/precip".format(target_path.replace("s3://", "")),
        ]
        timeout = 60 * 5
        start = time.time()
        print("[ RUNNING ]: ", " ".join(cmd))
        while True:
            proc = subprocess.run(cmd, capture_output=True, text=True)
            # purposely don't check proc.returncode since files might not exist yet

            # --prune prunes to two time steps by default, so we expect 2 time steps here
            # but four overall files:
            #
            # $ mc ls myminio/gpcp/target/recipe/gpcp/precip/
            # [2023-10-24 22:42:16 UTC]   365B STANDARD .zarray
            # [2023-10-24 22:42:16 UTC]   442B STANDARD .zattrs
            # [2023-10-24 22:42:17 UTC] 145KiB STANDARD 0.0.0
            # [2023-10-24 22:42:17 UTC] 148KiB STANDARD 1.0.0
            try:
                output = proc.stdout.splitlines()
                print(f"[ MINIO OUTPUT ]: {output[-1]}")
                if len(output) == 4:
                    break
            except:
                pass

            elapsed_time = time.time() - start
            if elapsed_time >= timeout:
                raise Exception("timeout reached, exiting")
            time.sleep(2)

        gpcp = xr.open_dataset(
            target_path, backend_kwargs={"storage_options": fsspec_args}, engine="zarr"
        )

        assert (
            gpcp.title
            == "Global Precipitation Climatatology Project (GPCP) Climate Data Record (CDR), Daily V1.3"
        )
        # --prune prunes to two time steps by default, so we expect 2 items here
        assert len(gpcp.precip) == 2
        print(gpcp)
